import os
import sys
import datetime
import traceback
import platform
from pydoc import locate

import imviz as viz
import minireload as mr


def launch(cls, func_name):

    file_path = os.path.abspath(sys.modules[cls.__module__].__file__)

    cls_name = os.path.basename(file_path).rsplit(".")[0]
    cls_name += "." + cls.__qualname__

    os.environ["PYTHONPATH"] = ":".join(
            sys.path + [os.path.dirname(file_path)])

    if platform.system() == 'Windows':
        # execlpe has different semantics on Windows
        cls = locate(cls_name)
        loop(cls, func_name)
    else:
        os.execlpe(sys.executable,
               "python3",
               "-m",
               "imviz.dev_main",
               cls_name,
               func_name,
               *sys.argv[1:],
               os.environ)


def try_load_source(path):

    try:
        with open(path) as fd:
            source = fd.readlines()
        return source
    except Exception:
        return ""


def build_launcher(cls, func_name):

    class ImvizLauncher:

        def __init__(self):

            viz.configure_ini_path(sys.modules[cls.__module__])

            self.obj = cls()
            self.func = getattr(self.obj, func_name)

            self.exc_info = None
            self.exc_frame_idx = 0
            self.exc_code = None
            self.request_scroll = False

        def loop(self):

            self.func()

        def exc_func(self, ei):

            if self.exc_info != ei:
                self.request_scroll = True
                self.exc_info = ei
                self.exc_frame_idx = max(0, len(ei.exc_frames) - 1)

                if isinstance(ei.exc, SyntaxError):
                    self.exc_code = try_load_source(ei.exc.filename)
                else:
                    self.exc_code = try_load_source(
                            ei.exc_frames[-1].filename)

            if not viz.wait():
                sys.exit()

            cx, cy = viz.get_viewport_center()
            w, h = viz.get_main_window_size()
            w *= 0.95
            h *= 0.95

            if viz.begin_window("Exception"):
                fade = min(1.0, (
                    datetime.datetime.now().timestamp()
                    - ei.exc_time.timestamp()) / 1.0)
                col = (1.0, fade, fade)

                time_string = ei.exc_time.strftime("%H:%M:%S")

                if viz.button("Reload"):
                    return True

                viz.separator()

                viz.text(f"Exception time {time_string}\n", color=col)
                viz.text(ei.exc_str, color=col)
                viz.text("\n")

                viz.separator()

                for i, f in enumerate(ei.exc_frames):
                    if viz.selectable(
                            f"{os.path.basename(f.filename)}"
                            + f" at line {f.lineno} in {f.function}",
                            i == self.exc_frame_idx):
                        self.exc_frame_idx = i
                        self.exc_code = try_load_source(f.filename)
                        self.request_scroll = True

            viz.end_window()

            if viz.begin_window("App state"):
                viz.autogui(self.obj.__dict__)
            viz.end_window()

            if isinstance(ei.exc_value, SyntaxError):
                viz.begin_window("Local variables")
                viz.end_window()

                self.render_source(
                        self.exc_code,
                        ei.exc_value.lineno-1)
            else:
                if viz.begin_window("Local variables"):
                    f_locals = ei.exc_frames[self.exc_frame_idx].frame.f_locals
                    try:
                        viz.AutoguiContext(ignore_custom=True).render(f_locals)
                    except:
                        viz.text("Could not render local variables!", color="red")
                viz.end_window()

                self.render_source(
                        self.exc_code,
                        ei.exc_frames[self.exc_frame_idx].lineno-1)

            return False

        def render_source(self, code_str, error_line):

            if viz.begin_window("Source code"):
                if viz.begin_table("code", 2,
                                   viz.TableFlags.BORDERS_INNER_V):

                    viz.table_setup_column(
                            "line_numbers",
                            viz.TableColumnFlags.WIDTH_FIXED
                            | viz.TableColumnFlags.NO_RESIZE)

                    viz.table_setup_column(
                            "source",
                            viz.TableColumnFlags.WIDTH_STRETCH)

                    for i, l in enumerate(code_str):
                        if i == error_line:
                            lineno_color = (1.0, 0.3, 0.3)
                            source_color = (1.0, 0.3, 0.3)
                            if self.request_scroll:
                                viz.set_scroll_here_y(0.5)
                                self.request_scroll = False
                        else:
                            lineno_color = (0.4, 0.4, 0.4)
                            source_color = (1.0, 1.0, 1.0)

                        viz.table_next_column()
                        viz.text(str(i+1), color=lineno_color)
                        viz.table_next_column()
                        viz.text(l, color=source_color)
                        viz.table_next_row()
                    viz.end_table()
            viz.end_window()

    return ImvizLauncher



"""
This contains common and (mostly) helpful utils.
"""

import os
import time
import pickle
import inspect
import hashlib
import traceback
import subprocess

import __main__

import imviz as viz

from contextlib import contextmanager
from minireload import ModuleReloader
from objtoolbox import bundle


STATICS = {}
"""
Contains all static function variables.
"""


def statics(**defaults):
    """
    This (kind of) emulates the behavior of c++ static function variables.

    It returns a bundle, which is unique to the calling function.
    Like static variables the bundle is persisted between function calls.

    Use with caution! This implementation is probably not very performant,
    but really useful for quick-and-dirty experimentation.
    """

    caller = inspect.stack()[1]
    func_id = hash(caller.filename + caller.function)

    try:
        sts = STATICS[func_id]
        sts = bundle({**defaults, **sts})
        STATICS[func_id] = sts
    except KeyError:
        sts = bundle(defaults)
        STATICS[func_id] = sts

    return sts


RELOADER = None
"""
Contains a global module reloader for easier access.
"""


def update_autoreload():
    """
    This checks all used python modules for changes (mtime), reloads
    the respective code, and updates existing code as far as possible.

    The "__main__" module cannot be reloaded due to python limitations.

    As this uses asynchronous modification scanning, the function
    must be called repeatedly in the main application loop.

    Returns True if at least one module was sucessfully reloaded.
    Returns False otherwise.
    """

    global RELOADER

    if RELOADER is None:
        RELOADER = ModuleReloader()

    return RELOADER.reload()


@contextmanager
def error_sink():
    """
    This contextmanager catches and visualizes exceptions in the gui,
    instead of handing them over to the caller.
    """

    try:
        yield
    except Exception as e:
        viz.text(f"{type(e).__name__}: {e}", color=(1, 0, 0))
        if viz.is_item_hovered():
            viz.begin_tooltip()
            viz.text(f"{traceback.format_exc(-1)}")
            viz.end_tooltip()


AUTOSAVE_REQ = {}
AUTOSAVE_TIME = {}


@contextmanager
def autosave(obj, path=".imviz_save", timeout=0.5):

    if path not in AUTOSAVE_REQ:
        AUTOSAVE_REQ[path] = False
    if path not in AUTOSAVE_TIME:
        AUTOSAVE_TIME[path] = -1

    if AUTOSAVE_TIME[path] < 0:
        viz.storage.load(obj, path)
        AUTOSAVE_TIME[path] = time.time()

    viz.push_mod_any()

    yield

    if viz.pop_mod_any():
        AUTOSAVE_REQ[path] = True
        AUTOSAVE_TIME[path] = time.time()

    if AUTOSAVE_REQ[path] and (time.time() - AUTOSAVE_TIME[path]) > timeout:
        AUTOSAVE_REQ[path] = False
        viz.storage.save(obj, path)


class Selection():
    """
    This class combines a combo box selection with a list of options.
    """

    def __init__(self, options=[], index=0):

        self.options = options
        self.index = 0

    def __autogui__(self, name, **kwargs):

        self.index = viz.combo(name, self.options, self.index)
        viz.same_line()
        self.options = viz.autogui(
                self.options, f"###{name}_options")

        return self

    def selected(self):

        try:
            return self.options[self.index]
        except IndexError:
            return None


class ModHistory:

    mod_counter = 0

    def __init__(self):

        self.pos = -1
        self.history = []
        self.time = 0
        self.save_req = False

    def save(self, obj):

        ModHistory.mod_counter += 1
        if self.pos < len(self.history) - 1:
            self.history = self.history[:self.pos - len(self.history) + 1]
        self.history.append((ModHistory.mod_counter, pickle.dumps(obj)))
        self.pos += 1

    def get_undo_id(self):

        if self.pos - 1 >= 0:
            return self.history[self.pos][0]
        else:
            return 0

    def get_undo_state(self):

        if self.pos - 1 >= 0:
            return pickle.loads(self.history[self.pos-1][1])
        else:
            return None

    def get_redo_id(self):

        if self.pos + 1 <= len(self.history) - 1:
            return self.history[self.pos][0]
        else:
            return 0

    def get_redo_state(self):

        if self.pos + 1 <= len(self.history) - 1:
            return pickle.loads(self.history[self.pos+1][1])
        else:
            return None


MOD_HISTORIES = {}
UNDO_CANDIDATE = None
REDO_CANDIDATE = None


@contextmanager
def mod_history(name, obj, timeout=0.5):

    global UNDO_CANDIDATE
    global REDO_CANDIDATE

    hist_id = viz.get_id(name)

    try:
        hist = MOD_HISTORIES[hist_id]
    except KeyError:
        hist = ModHistory()
        hist.save(obj)
        MOD_HISTORIES[hist_id] = hist

    if UNDO_CANDIDATE == hist:
        obj.__dict__ = hist.get_undo_state().__dict__
        hist.pos -= 1
        UNDO_CANDIDATE = None
        viz.set_mod(True)

    if REDO_CANDIDATE == hist:
        obj.__dict__ = hist.get_redo_state().__dict__
        hist.pos += 1
        REDO_CANDIDATE = None
        viz.set_mod(True)

    for ke in viz.get_key_events():
        if ke.action == viz.RELEASE:
            continue
        if ke.mod != viz.MOD_CONTROL:
            continue
        if ke.key == viz.KEY_Y:
            if hist.get_redo_id() == 0:
                break
            if REDO_CANDIDATE is None:
                REDO_CANDIDATE = hist
                break
            if REDO_CANDIDATE.get_redo_id() > hist.get_redo_id():
                REDO_CANDIDATE = hist
                break
        if ke.key == viz.KEY_Z:
            if hist.get_undo_id() == 0:
                break
            if UNDO_CANDIDATE is None:
                UNDO_CANDIDATE = hist
                break
            if UNDO_CANDIDATE.get_undo_id() < hist.get_undo_id():
                UNDO_CANDIDATE = hist
                break

    viz.push_mod_any()

    yield

    if viz.pop_mod_any():
        hist.save_req = True
        hist.time = time.time()

    if hist.save_req and (time.time() - hist.time) > timeout:
        hist.save_req = False
        hist.save(obj)



"""
This contains functions to export plots in various formats.
"""


import io
import os
import sys
import time
import base64
import pickle
import subprocess
import numpy as np

from PIL import Image


pdf_avail = subprocess.call("which inkscape",
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE) == 0


try:
    import cppimviz as viz
except ModuleNotFoundError:
    sys.path.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "../build"))
    import cppimviz as viz


from imviz.common import bundle


class Vertex:

    def __init__(self):

        self.pos = np.array([0.0, 0.0])
        self.uv = np.array([0.0, 0.0])

    def __eq__(self, o):

        return ((self.pos == o.pos).all()
                and (self.uv == o.uv).all())

    def __hash__(self):

        return hash((*self.pos, *self.uv))


class Polygon:

    def __init__(self):

        self.color = ""
        self.alpha = 0.0
        self.vertices = []

        # only used, if the polygon represents one or many characters

        self.text = ""
        self.font_size = 0.0
        self.advance = 0.0
        self.last_char_x = 0.0
        self.last_char_y = 0.0
        self.vertical_text = False
        self.start_x = 0.0
        self.start_y = 0.0
        self.x0 = 0.0
        self.y0 = 0.0
        self.x1 = 0.0
        self.y1 = 0.0

        # only use if we have an image in the polygon

        self.image = None


class Polyline:

    def __init__(self):

        self.points = []
        self.color = ""
        self.alpha = 0.0


class DrawListState:

    def __init__(self):

        self.draw_cmds = []
        self.polygon_groups = []
        self.line_groups = []

        self.canvas_pos = np.array([0.0, 0.0])
        self.canvas_size = np.array([0.0, 0.0])


def export_polygons(state, dl):

    idxs = dl.get_indices()
    verts = dl.get_verts()

    for cmd in state.draw_cmds:

        polygons = []
        merged_polys = []

        # convert into polygons

        start = cmd.idx_offset
        end = cmd.idx_offset + cmd.elem_count

        for i0, i1, i2 in zip(idxs[start:end:3],
                              idxs[start+1:end:3],
                              idxs[start+2:end:3]):

            c0 = verts[i0].col
            c1 = verts[i1].col
            c2 = verts[i2].col

            alpha0 = ((c0 & 0xff000000) >> 24) / 255
            alpha1 = ((c1 & 0xff000000) >> 24) / 255
            alpha2 = ((c2 & 0xff000000) >> 24) / 255

            alpha = round(max([alpha0, alpha1, alpha2]), 3)

            blue = (c0 & 0x00ff0000) >> 16
            green = (c0 & 0x0000ff00) >> 8
            red = c0 & 0x000000ff

            v0 = Vertex()
            v0.pos = verts[i0].pos
            v0.uv = verts[i0].uv

            v1 = Vertex()
            v1.pos = verts[i1].pos
            v1.uv = verts[i1].uv

            v2 = Vertex()
            v2.pos = verts[i2].pos
            v2.uv = verts[i2].uv

            p = Polygon()
            p.vertices = [v0, v1, v2]
            p.color = f"#{red:02x}{green:02x}{blue:02x}"
            p.alpha = alpha

            polygons.append(p)

        # merge polygons

        if len(polygons) > 0:

            p = polygons[0]

            for o in polygons[1:]:

                if p.color != o.color:
                    merged_polys.append(p)
                    p = o
                    continue

                intersect = list(set(o.vertices) & set(p.vertices))

                if len(intersect) != 2:
                    merged_polys.append(p)
                    p = o
                    continue

                # detected matching side, joining

                p_i0 = p.vertices.index(intersect[0])
                p_i1 = p.vertices.index(intersect[1])
                if p_i1 < p_i0:
                    p_i1, p_i0 = p_i0, p_i1

                excluded_vtxs = list(set(o.vertices) - set(intersect))

                if len(excluded_vtxs) != 1:
                    merged_polys.append(p)
                    p = o
                    continue

                excluded_vtxs = excluded_vtxs[0]

                if p_i1 - p_i0 == 1:
                    p.vertices.insert(p_i0+1, excluded_vtxs)
                else:
                    p.vertices.append(excluded_vtxs)

            merged_polys.append(p)

        state.polygon_groups.append(merged_polys)


def merge_polygons_to_lines(state):

    def get_mid_points(p):

        axis_0 = (p.vertices[3].pos - p.vertices[0].pos) / 2.0
        axis_1 = (p.vertices[1].pos - p.vertices[0].pos) / 2.0
        center = p.vertices[0].pos + axis_0 + axis_1

        mp0 = center + axis_0 
        mp1 = center + axis_1
        mp2 = center - axis_0 
        mp3 = center - axis_1

        return [mp0, mp1, mp2, mp3]

    for pg in state.polygon_groups:

        prev_p = None

        line_group = []

        line_points = []
        remove_list = []
        counter_point = None

        for p in pg:

            if len(p.vertices) != 4:
                prev_p = None 
                continue

            if prev_p is None:
                prev_p = p
                continue

            p_mid_points = get_mid_points(p)
            pp_mid_points = get_mid_points(prev_p)

            if prev_p.color != p.color:
                if len(line_points) > 0:
                    line_points.append(counter_point)
                    pl = Polyline()
                    pl.color = remove_list[-1].color
                    pl.alpha = remove_list[-1].alpha
                    pl.points = line_points
                    line_group.append(pl)
                    line_points = []
                prev_p = None
                continue

            if prev_p.alpha != p.alpha:
                if len(line_points) > 0:
                    line_points.append(counter_point)
                    pl = Polyline()
                    pl.color = remove_list[-1].color
                    pl.alpha = remove_list[-1].alpha
                    pl.points = line_points
                    line_group.append(pl)
                    line_points = []
                prev_p = None
                continue

            matched = False

            for i, mp in enumerate(p_mid_points):
                for j, mpp in enumerate(pp_mid_points):
                    if np.linalg.norm(mp - mpp) < 1.0:
                        matched = True
                        break

            if matched:
                if len(line_points) == 0:
                    line_points.append(pp_mid_points[(j + 2) % 4])
                    remove_list.append(prev_p)
                line_points.append(mp)
                remove_list.append(p)
                counter_point = p_mid_points[(i + 2) % 4]
            elif len(line_points) > 0:
                line_points.append(counter_point)
                pl = Polyline()
                pl.color = remove_list[-1].color
                pl.alpha = remove_list[-1].alpha
                pl.points = line_points
                line_group.append(pl)
                line_points = []

            prev_p = p

        for p in remove_list:
            pg.remove(p)

        if len(line_points) > 0:

            line_points.append(p_mid_points[(i + 2) % 4])
            pl = Polyline()
            pl.color = remove_list[-1].color
            pl.alpha = remove_list[-1].alpha
            pl.points = line_points
            line_group.append(pl)
            line_points = []

        state.line_groups.append(line_group)


def export_text_polygons(state):

    # build a character lookup table based on
    # uv coordinates of the font atlas texture

    fonts = viz.get_font_atlas().get_fonts()
    uv_to_char = {}

    for font in fonts:
        for g in font.get_glyphs():
            k = str(g.u0) + str(g.v0)
            c = bytearray([g.codepoint, 0]).decode("utf16")

            v = bundle()
            v.text = c
            v.font_size = font.font_size
            v.advance = g.advance_x
            v.x0 = g.x0
            v.y0 = g.y0
            v.x1 = g.x1
            v.y1 = g.y1

            uv_to_char[k] = v

    # iterate all groups to identify characters

    txt_tex_id = viz.get_font_atlas().get_texture_id()

    pp = None

    for i, (cmd, polys) in enumerate(
            zip(state.draw_cmds, state.polygon_groups)):

        if txt_tex_id != cmd.texture_id:
            continue

        new_polys = []
        half_space = False

        for p in polys:

            if len(p.vertices) != 4:
                new_polys.append(p)
                continue

            if (p.vertices[0].uv == p.vertices[1].uv).all():
                new_polys.append(p)
                continue

            # found something with a font texture

            uv0 = np.min(np.array(
                [v.uv for v in p.vertices]), axis=0)

            k = str(float(uv0[0])) + str(float(uv0[1]))

            try:
                char_info = uv_to_char[k]
            except KeyError:
                new_polys.append(p)
                continue

            # seems to be a character

            p.text = char_info.text
            p.font_size = char_info.font_size
            p.advance = char_info.advance
            p.x0 = char_info.x0
            p.y0 = char_info.y0
            p.x1 = char_info.x1
            p.y1 = char_info.y1

            # check if we have vertical text

            min_vtx = min(p.vertices, key=lambda v: v.pos[0]**2 + v.pos[1]**2)
            min_uv = min(p.vertices, key=lambda v: v.uv[0]**2 + v.uv[1]**2)

            p.vertical_text = min_vtx is not min_uv

            # determine position

            min_x, min_y = np.array([v.pos for v in p.vertices]).min(axis=0)
            max_x, max_y = np.array([v.pos for v in p.vertices]).max(axis=0)

            if p.vertical_text:
                p.last_char_x = min_x
                p.last_char_y = max_y
                p.start_x = max_x - p.y1 + p.font_size/2
                p.start_y = max_y + p.x0
            else:
                p.last_char_x = min_x
                p.last_char_y = min_y
                p.start_x = min_x - p.x0
                p.start_y = min_y - p.y0 + p.font_size/2

            # check if we can merge it with a previous char

            if len(new_polys) == 0:
                new_polys.append(p)
                pp = new_polys[-1]
                continue

            if pp is None:
                new_polys.append(p)
                pp = new_polys[-1]
                continue

            if p.vertical_text:
                adv_dist = (pp.last_char_y - p.x0) - (p.last_char_y - pp.x0)
                base_dist = (p.last_char_x - p.y0) - (pp.last_char_x - pp.y0)
            else:
                adv_dist = (p.last_char_x - p.x0) - (pp.last_char_x - pp.x0)
                base_dist = (p.last_char_y - p.y0) - (pp.last_char_y - pp.y0)

            space_size = 3.1821829676628113
            space_step = abs(adv_dist - pp.advance) / space_size
            round_space_steps = round(space_step)

            if abs(space_step - round_space_steps) < 1e-3:
                pp.text += " " * round_space_steps

            can_be_joined = (abs(adv_dist - pp.advance) < p.font_size
                             and abs(base_dist) < 1e-4
                             and (pp.font_size == p.font_size)
                             and (pp.color == p.color)
                             and (pp.alpha == p.alpha)
                             and (pp.vertical_text == p.vertical_text))

            if not can_be_joined:
                new_polys.append(p)
                pp = new_polys[-1]
                continue

            # join and continue

            pp.text += p.text
            pp.vertices += p.vertices
            pp.advance = p.advance
            pp.last_char_x = p.last_char_x
            pp.last_char_y = p.last_char_y
            pp.x0 = p.x0
            pp.y0 = p.y0

        state.polygon_groups[i] = new_polys


def export_images(state):

    txt_tex_id = viz.get_font_atlas().get_texture_id()

    for cmd, pg in zip(state.draw_cmds, state.polygon_groups):

        if cmd.texture_id == txt_tex_id:
            continue

        # found something with a different texture
        # we need to get that texture!

        texture = viz.get_texture(cmd.texture_id)

        for p in pg:
            p.image = texture


def export_canvas(state):

    canvas_min = np.ones(2) * float("inf")
    canvas_max = np.ones(2) * -float("inf")

    for cmd in state.draw_cmds:
        canvas_min[0] = min(canvas_min[0], cmd.clip_rect[0])
        canvas_min[1] = min(canvas_min[1], cmd.clip_rect[1])
        canvas_max[0] = max(canvas_max[0], cmd.clip_rect[2])
        canvas_max[1] = max(canvas_max[1], cmd.clip_rect[3])

    state.canvas_pos = canvas_min
    state.canvas_size = canvas_max - canvas_min


def export_drawlist_state(dl):

    state = DrawListState()
    state.draw_cmds = dl.get_cmds()

    export_polygons(state, dl)
    export_text_polygons(state)
    # merge_polygons_to_lines(state)
    export_images(state)
    export_canvas(state)

    return state


def polygon_to_svg(p):

    if p.image is not None:

        np_vtx = np.array([v.pos for v in p.vertices])
        box_min = np_vtx.min(axis=0)
        box_max = np_vtx.max(axis=0)
        box_dims = box_max - box_min

        with io.BytesIO() as fd:
            img = Image.fromarray(p.image)
            img.save(fd, "png")
            fd.seek(0)
            str_img = base64.b64encode(fd.read()).decode("utf8")

        svg_txt = f'<image x="{box_min[0]:.3f}" y="{box_min[1]:.3f}" '
        svg_txt += f'width="{box_dims[0]:.3f}" height="{box_dims[1]:.3f}" '
        svg_txt += 'preserveAspectRatio="none" '
        svg_txt += f'xlink:href="data:image/png;base64,{str_img}" '
        svg_txt += '/>'

        return svg_txt

    if len(p.text) == 0:
        svg_txt = '<polygon points="'

        for i, v in enumerate(p.vertices):
            svg_txt += f'{v.pos[0]:.3f},{v.pos[1]:.3f}'
            if i != len(p.vertices) - 1:
                svg_txt += " "

        svg_txt += f'" fill="{p.color}" fill-opacity="{p.alpha}" />'
    else:
        # no idea why this is necessary, but it works
        char_size = p.font_size * 0.8

        np_vtx = np.array([v.pos for v in p.vertices])
        box_min = np_vtx.min(axis=0)
        box_max = np_vtx.max(axis=0)
        box_mean = box_min + (box_max - box_min) * 0.5

        t_x, t_y = box_mean[0], box_mean[1]

        if p.vertical_text:
            svg_txt = ('<text '
                       + 'transform="translate('
                       + f'{(t_x + 5):.3f}, {t_y:.3f}) '
                       + 'rotate(-90)" ')
        else:
            svg_txt = f'<text x="{t_x:.3f}" y="{(t_y + 5):.3f}" '

        svg_txt += (f'fill="{p.color}" fill-opacity="{p.alpha}" '
                    + 'style="font-family: Source Sans Pro; '
                    + f'font-size: {char_size}px; '
                    + f'text-anchor: middle;" '
                    + f'>{p.text}</text>')

    return svg_txt


def line_to_svg(ln):

    svg_txt = '<polyline points="'
    for p in ln.points:
        svg_txt += f'{p[0]:.3f}, {p[1]:.3f} '
    svg_txt += f'" fill="none" stroke="{ln.color}" stroke-opacity="{ln.alpha}" />'

    return svg_txt


def drawlist_state_to_svg(state):

    # output svg text

    svg_txt = ('<svg version="1.1" '
               + 'viewBox="'
               + f'{state.canvas_pos[0]} '
               + f'{state.canvas_pos[1]} '
               + f'{state.canvas_size[0]} '
               + f'{state.canvas_size[1]}" '
               + 'xmlns="http://www.w3.org/2000/svg" '
               + 'xmlns:xlink="http://www.w3.org/1999/xlink">\n')

    # write clip rects

    svg_txt += "<defs>\n"

    for i, cmd in enumerate(state.draw_cmds):
        cx = cmd.clip_rect[0]
        cy = cmd.clip_rect[1]
        cw = cmd.clip_rect[2] - cx
        ch = cmd.clip_rect[3] - cy
        svg_txt += (f'<clipPath id="clip_rect_{i}">'
                    + f'<rect x="{cx}" y="{cy}" '
                    + f'width="{cw}" height="{ch}" />'
                    + '</clipPath>\n')

    svg_txt += "</defs>\n"

    # write out polygon and line groups

    for i, pg in enumerate(state.polygon_groups):
        svg_txt += f'<g clip-path="url(#clip_rect_{i})">\n'
        for p in pg:
            svg_txt += polygon_to_svg(p) + "\n"
        svg_txt += '</g>\n'

        #svg_txt += f'<g clip-path="url(#clip_rect_{i})">\n'
        #for ln in state.line_groups[i]:
        #    svg_txt += line_to_svg(ln) + "\n"
        #svg_txt += '</g>\n'

    # close and return

    svg_txt += '</svg>\n'

    return svg_txt


class PlotExport:

    plot_id = -1
    countdown = 0
    filetype = ""
    path = os.path.abspath(os.getcwd())
    csv_data = {}


def wrap_begin(begin_func):

    def inner(*args, **kwargs):

        res = begin_func(*args, **kwargs)

        if (PlotExport.plot_id == viz.get_plot_id()
                and PlotExport.countdown < 1
                and PlotExport.filetype != "png"):
            viz.disable_aa()
            if PlotExport.filetype == "csv":
                csv_data = {}

        return res

    return inner


def wrap_plot(plot_func):

    def inner(*args, **kwargs):

        plot_func(*args, **kwargs)

        if (PlotExport.plot_id == viz.get_plot_id()
                and PlotExport.countdown < 1
                and PlotExport.filetype == "csv"):
            if "label" in kwargs:
                data_key = kwargs["label"]
            else:
                data_key = str(len(PlotExport.csv_data))

            if len(args) == 1:
                y_data = np.array(args[0])
                x_data = np.arange(len(args[0]))
            elif len(args) == 2:
                y_data = np.array(args[1])
                x_data = np.array(args[0])

            l = min(x_data.shape[0], y_data.shape[0])
            PlotExport.csv_data[data_key] = np.vstack((x_data[:l], y_data[:l])).T

    return inner


def wrap_end(end_func):

    def inner():

        current_plot_id = viz.get_plot_id()

        if current_plot_id is None:
            end_func()
            return

        dl = viz.get_window_drawlist()

        file_dialog_requested = False

        viz.push_override_id(current_plot_id)
        if viz.begin_popup("##PlotContext"):
            if viz.begin_menu("Export"):
                if viz.menu_item("As csv"):
                    PlotExport.filetype = "csv"
                    file_dialog_requested = True
                if viz.menu_item("As pdf", enabled=pdf_avail):
                    PlotExport.filetype = "pdf"
                    file_dialog_requested = True
                if viz.menu_item("As pdf + latex", enabled=pdf_avail):
                    PlotExport.filetype = "pdf_tex"
                    file_dialog_requested = True
                if viz.menu_item("As png"):
                    PlotExport.filetype = "png"
                    file_dialog_requested = True
                if viz.menu_item("As svg"):
                    PlotExport.filetype = "svg"
                    file_dialog_requested = True
                viz.end_menu()
            viz.separator()
            viz.end_popup()

        if file_dialog_requested:
            viz.open_popup("Select export path")

        # create export path chooser

        PlotExport.path = viz.file_dialog_popup(
                "Select export path",
                PlotExport.path,
                "Export")

        export_requested = viz.mod()

        viz.pop_id()

        plot_pos = viz.get_window_pos()
        plot_size = viz.get_window_size()

        end_func()

        # do actual export

        if PlotExport.plot_id == current_plot_id:
            if PlotExport.countdown < 1:

                os.makedirs(PlotExport.path, exist_ok=True)

                if os.path.isdir(PlotExport.path):
                    PlotExport.path += "/"
                if PlotExport.filetype != "csv" and PlotExport.path.endswith("/"):
                    PlotExport.path += f"plot_{int(time.time() * 10**9)}"

                if PlotExport.filetype == "csv":

                    for k, v in PlotExport.csv_data.items():
                        np.savetxt(os.path.join(PlotExport.path, f"{k}.csv"),
                                   v,
                                   delimiter=",",
                                   header="x,y",
                                   comments="")

                elif PlotExport.filetype == "svg":

                    if not PlotExport.path.endswith(".svg"):
                        PlotExport.path += ".svg"

                    dl_state = export_drawlist_state(dl)
                    svg_txt = drawlist_state_to_svg(dl_state)

                    with open(PlotExport.path, "w+") as fd:
                        fd.write(svg_txt)

                elif PlotExport.filetype == "png":

                    if not PlotExport.path.endswith(".png"):
                        PlotExport.path += ".png"

                    pixels = viz.get_pixels(plot_pos[0],
                                            plot_pos[1],
                                            plot_size[0],
                                            plot_size[1])

                    Image.fromarray(pixels).save(PlotExport.path)

                elif PlotExport.filetype == "pdf":

                    # first export as svg

                    tmp_path = f"/tmp/imviz_exp_{int(time.time() * 10**9)}.svg"

                    dl_state = export_drawlist_state(dl)
                    svg_txt = drawlist_state_to_svg(dl_state)

                    with open(tmp_path, "w+") as fd:
                        fd.write(svg_txt)

                    # then convert to pdf

                    if not PlotExport.path.endswith(".pdf"):
                        PlotExport.path += ".pdf"

                    exp_cmd = (f'inkscape --file="{tmp_path}" '
                               + '--export-area-drawing '
                               + '--without-gui '
                               + f'--export-pdf="{PlotExport.path}"')

                    subprocess.call(exp_cmd,
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)

                elif PlotExport.filetype == "pdf_tex":

                    # first export as svg

                    tmp_path = f"/tmp/imviz_exp_{int(time.time() * 10**9)}.svg"

                    dl_state = export_drawlist_state(dl)
                    svg_txt = drawlist_state_to_svg(dl_state)

                    with open(tmp_path, "w+") as fd:
                        fd.write(svg_txt)

                    # then convert to pdf

                    if not PlotExport.path.endswith(".pdf"):
                        PlotExport.path += ".pdf"

                    exp_cmd = (f'inkscape --file="{tmp_path}" '
                               + '--export-area-drawing '
                               + '--without-gui '
                               + '--export-latex '
                               + f'--export-pdf="{PlotExport.path}"')

                    subprocess.call(exp_cmd,
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)

                PlotExport.plot_id = -1
                PlotExport.filetype = ""

            else:
                PlotExport.countdown -= 1

        if export_requested:
            PlotExport.plot_id = current_plot_id
            PlotExport.countdown += 2

    return inner


begin_plot = wrap_begin(viz.begin_plot)
begin_figure = wrap_begin(viz.begin_figure)

plot = wrap_plot(viz.plot)

end_plot = wrap_end(viz.end_plot)
end_figure = wrap_end(viz.end_figure)



from concurrent.futures import ThreadPoolExecutor


THREAD_POOL = ThreadPoolExecutor(32)
TASK_FUTURES = {}


def start(tid, func, *args, **kwargs):
    """
    This calls the given function in a background thread.
    Multiple calls with the same tid will interrupt already running tasks.
    """

    try:
        task_future = TASK_FUTURES[tid]
        if task_future is not None:
            task_future.cancel()
    except KeyError:
        pass

    TASK_FUTURES[tid] = THREAD_POOL.submit(func, *args, **kwargs)


def update(tid, func, *args, **kwargs):
    """
    This calls the given function in a background thread.
    Multiple calls with the same tid will not queue tasks.

    If the task is currently running, no new task will be started,
    nor will the running task be interrupted.
    """

    try:
        task_future = TASK_FUTURES[tid]
    except KeyError:
        task_future = None

    if task_future is None:
        TASK_FUTURES[tid] = THREAD_POOL.submit(func, *args, **kwargs)


def result(tid):
    """
    This returns the task result and resets the task state to "inactive".
    Returns None if no task result is present.
    """

    try:
        task_future = TASK_FUTURES[tid]
        if task_future is None:
            return None
        if not task_future.done():
            return None

        TASK_FUTURES[tid] = None
        return task_future.result()
    except KeyError:
        return None


def cancel(tid):

    try:
        task_future = TASK_FUTURES[tid]
        task_future.cancel()
        TASK_FUTURES[tid] = None
    except KeyError:
        pass


def active(tid):

    try:
        task_future = TASK_FUTURES[tid]
        if task_future is None:
            return False
        else:
            return not task_future.done()
    except KeyError:
        return False



import os
import sys


try:
    from cppimviz import *
except ModuleNotFoundError:
    sys.path.append(os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "../build"))
    from cppimviz import *


INITIAL_INI = """
[Window][Exception]
Pos=0,210
Size=400,400
Collapsed=0
DockId=0x00000005,0

[Window][App state]
Pos=399,0
Size=400,600
Collapsed=0
DockId=0x00000003,1

[Window][Local variables]
Pos=0,0
Size=400,200
Collapsed=0
DockId=0x00000004,0

[Window][Source code]
Pos=399,0
Size=400,600
Collapsed=0
DockId=0x00000003,0

[Docking][Data]
DockSpace     ID=0x00000001 Window=0x70D0451F Pos=0,0 Size=800,600 Split=X Selected=0x8B814D52
  DockNode    ID=0x00000002 Parent=0x00000001 SizeRef=400,600 Split=Y Selected=0xFD380CCE
    DockNode  ID=0x00000004 Parent=0x00000002 SizeRef=400,200 Selected=0xF90B12C6
    DockNode  ID=0x00000005 Parent=0x00000002 SizeRef=400,400 Selected=0xFD380CCE
  DockNode    ID=0x00000003 Parent=0x00000001 SizeRef=400,600 CentralNode=1 Selected=0x8B814D52
"""


def configure_ini_path(module):

    if not hasattr(module, "__file__"):
        return

    main_file_name = os.path.basename(module.__file__).rsplit(".")[0]
    ini_path = os.path.join(
            os.path.abspath(os.path.dirname(module.__file__)),
            "." + main_file_name + ".imviz.ini")

    set_ini_path(ini_path)
    load_ini_from_str(INITIAL_INI)
    load_ini(ini_path)


import __main__
configure_ini_path(__main__)


from imviz.common import *
from imviz.autogui import autogui_func as autogui
from imviz.autogui import AutoguiContext

import imviz.dev
import imviz.task


from imviz.export import (begin_plot, end_plot, plot, begin_figure, end_figure)



"""
Because the automatic code reloading cannot reload the __main__ module,
we provide a module here, which can launch and reload a given class.

This frees the library user from providing such a main module themselves.

For specific needs not covered by this simple implementation, you may use
this module as a reference for your own (development) main module.
"""

import argparse
import minireload as mr

# i like this
from pydoc import locate

from imviz.dev import build_launcher


def main():

    parser = argparse.ArgumentParser(
            description="Launch a class with automatic code reloading")

    parser.add_argument(
            "class_name",
            type=str,
            help="the name of the class to instantiate")

    parser.add_argument(
            "func_name",
            type=str,
            help="the name of the method to call")

    args, unknown = parser.parse_known_args()

    cls = locate(args.class_name)

    if cls is None:
        print(f"Could not find class {args.class_name}")
        return

    imviz_launcher = build_launcher(cls, args.func_name)

    mr.loop(imviz_launcher, "loop", "exc_func")


if __name__ == "__main__":
    main()



import copy
import typing
import numbers
import itertools
import traceback

import imviz as viz
import numpy as np

from imviz.storage import ext_setattr


def autogui_func(obj, name="", **params):

    return AutoguiContext(params=params).render(obj, name)


def list_item_context(obj, name, ctx):

    i = ctx.path[-1]

    if viz.begin_popup_context_item():
        if hasattr(ctx.parents[-1], "insert"):
            if viz.menu_item("Duplicate"):
                ctx.duplicate_item = i
        if hasattr(ctx.parents[-1], "__delitem__"):
            if viz.menu_item("Remove"):
                ctx.remove_item = i
                viz.set_mod(True)
        viz.end_popup()

    ctx.post_header_hooks.remove(list_item_context)


class AutoguiContext:

    def __init__(self,
                 params={},
                 path=[],
                 parents=[],
                 annotation=None,
                 ignore_custom=False):

        self.params = params
        self.path = path
        self.parents = parents
        self.annotation = annotation
        self.ignore_custom = ignore_custom

        self.post_header_hooks = []
        self.remove_item = None
        self.duplicate_item = None

        self.path_of_mod_item = []

    def try_render(self, obj, name=""):

        try:
            return self.try_render(obj, name)
        except Exception as e:
            viz.text(f"{name}: {e}")
            if viz.is_item_hovered():
                viz.begin_tooltip()
                viz.text(traceback.format_exc())
                viz.end_tooltip()

        return obj

    def call_post_header_hooks(self, obj, name):

        for h in self.post_header_hooks:
            h(obj, name, self)

    def render(self, obj, name=""):

        """
        This inspects the given object and renders a gui
        (with best effort) for all fields of the object.

        Rendering can be customized by defining an "__autogui__"
        method in the objects to be rendered.
        """

        obj_type = type(obj)

        if hasattr(obj, "__autogui__") and not self.ignore_custom:
            viz.push_mod_any()
            res = obj.__autogui__(name, ctx=self, **self.params)
            if viz.pop_mod_any():
                if self.path_of_mod_item == []:
                    self.path_of_mod_item = copy.deepcopy(self.path)
            return res

        if obj is None:
            viz.text(f"{name}: None")
            self.call_post_header_hooks(obj, name)
            return obj

        if obj_type == bool or obj_type == np.bool_:
            if name == "":
                name = str(self.path[-1])
            res = viz.checkbox(name, obj)
            self.call_post_header_hooks(obj, name)
            return res

        if isinstance(obj, numbers.Integral):
            if name == "":
                name = str(self.path[-1])
            res = obj_type(viz.drag(name, obj, 1.0, 0, 0))
            self.call_post_header_hooks(obj, name)
            return res

        if isinstance(obj, numbers.Real):
            if name == "":
                name = str(self.path[-1])
            res = obj_type(viz.drag(name, obj))
            self.call_post_header_hooks(obj, name)
            return res

        if obj_type == str:
            if name == "":
                name = str(self.path[-1])
            res = viz.input(name, obj)
            self.call_post_header_hooks(obj, name)
            return res

        # to reduce array lookups we collect the requested indices in "path"
        # the lookup then happens just before rendering the actual values
        indices = tuple(itertools.takewhile(
                lambda x: isinstance(x[0], int) and type(x[1]) == type(obj),
                    zip(self.path[::-1], self.parents[::-1])))[::-1]
        indices = tuple(i[0] for i in indices)
        li = len(indices)

        if len(name) > 0:
            tree_node_label = f"{name} "
            if hasattr(obj, "shape") and type(obj.shape) == tuple:
                tree_node_label += f"{list(obj.shape)[li:]}"
            elif hasattr(obj, "__len__"):
                tree_node_label += f"[{len(obj)}]"
            tree_node_label += f"###{name}"

            tree_open = viz.tree_node(tree_node_label)
            self.call_post_header_hooks(obj, name)
        else:
            tree_open = True

        if hasattr(obj, "shape") and hasattr(obj, "__getitem__"):

            if tree_open:
                if len(obj.shape) - li == 2:

                    # this tremendously speeds up zarr array access
                    # because we are now operating on a numpy array
                    arr_view = obj[indices]

                    width_avail, _ = viz.get_content_region_avail()
                    item_width = max(64, width_avail / arr_view.shape[-1] - 8)

                    for i in range(arr_view.shape[-2]):
                        for j in range(arr_view.shape[-1]):

                            viz.set_next_item_width(item_width)

                            self.path.append(i)
                            self.path.append(j)
                            self.parents.append(obj)

                            res = self.render(arr_view[i, j], f"###{i},{j}")

                            if viz.is_item_hovered():
                                viz.begin_tooltip()
                                viz.text(f"({i}, {j})")
                                viz.end_tooltip()

                            if viz.mod():
                                obj[indices + (i, j)] = res
                                if self.path_of_mod_item == []:
                                    self.path_of_mod_item = copy.deepcopy(self.path)

                            if j < arr_view.shape[-1] - 1:
                                viz.same_line()

                            self.parents.pop()
                            self.path.pop()
                            self.path.pop()

                elif len(obj.shape) - li == 1:

                    # lookup happens here
                    arr_view = obj[indices]

                    for i in range(len(arr_view)):

                        self.path.append(i)
                        self.parents.append(obj)

                        res = self.render(arr_view[i], str(i))

                        if viz.mod():
                            obj[indices + (i,)] = res
                            if self.path_of_mod_item == []:
                                self.path_of_mod_item = copy.deepcopy(self.path)

                        self.parents.pop()
                        self.path.pop()
                else:
                    for i in range(obj.shape[li]):

                        self.path.append(i)
                        self.parents.append(obj)

                        res = self.render(obj, str(i))

                        self.parents.pop()
                        self.path.pop()

            if len(name) > 0 and tree_open:
                viz.tree_pop()

            return obj

        if hasattr(obj, "__dict__"):
            attr_dict = obj.__dict__
        elif hasattr(obj, "__slots__"):
            attr_dict = {n: getattr(obj, n) for n in obj.__slots__}
        elif isinstance(obj, dict):
            attr_dict = obj
        elif hasattr(obj, "__len__") and hasattr(obj, "__getitem__"):
            if tree_open:
                for i in range(len(obj)):

                    self.post_header_hooks.append(list_item_context)

                    self.path.append(i)
                    self.parents.append(obj)

                    viz.push_mod_any()
                    res = self.render(obj[i], str(i))

                    if viz.pop_mod_any() and res is not obj[i]:
                        try:
                            obj.__setitem__(i, res)
                        except AttributeError:
                            pass
                        if self.path_of_mod_item == []:
                            self.path_of_mod_item = copy.deepcopy(self.path)

                    self.parents.pop()
                    self.path.pop()

                if self.duplicate_item is not None:
                    obj.insert(self.duplicate_item, copy.deepcopy(obj[self.duplicate_item]))
                    viz.set_mod(True)
                    self.duplicate_item = None

                if self.remove_item is not None:
                    del obj[self.remove_item]
                    viz.set_mod(True)
                    self.remove_item = None

            if len(name) > 0 and tree_open:
                viz.tree_pop()

            return obj
        else:
            if len(name) == 0:
                viz.text(f"{name}: " + "???")
            return obj

        obj_annots = getattr(obj, "__annotations__", {})
    
        # default case: generic object

        if tree_open:
            for k, v in attr_dict.items():

                if k in obj_annots:
                    self.annotation = obj_annots[k]

                self.path.append(k)
                self.parents.append(obj)

                viz.push_mod_any()
                new_v = self.render(v, str(k))

                if viz.pop_mod_any() and new_v is not v:
                    try:
                        ext_setattr(obj, k, new_v)
                    except AttributeError:
                        pass
                    if self.path_of_mod_item == []:
                        self.path_of_mod_item = copy.deepcopy(self.path)

                self.annotation = None
                self.parents.pop()
                self.path.pop()

        if len(name) > 0 and tree_open:
            viz.tree_pop()

        return obj



import os
import sys
import platform
import subprocess
import multiprocessing

from pathlib import Path
from typing import List

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install

# Setuptools calls cmake, which takes care of further
# C++ dependency resolution and the extension build.

# Inspired by:
# https://gist.github.com/hovren/5b62175731433c741d07ee6f482e2936
# https://www.benjack.io/2018/02/02/python-cpp-revisited.html
# Thanks a lot, very helpful :)

binary_files: List[Path] = []

class Install(install):
    def run(self):
        if self.root:
            # setup.py bdist_wheel
            os.makedirs(self.root, exist_ok=True)
            for f in binary_files:
                self.copy_file(f, Path(self.root) / f.name)
        else:
            # setup.py install
            pass
        
        # Continue as normal
        super().run()

class CMakeExtension(Extension):
    def __init__(self, name):
        Extension.__init__(self, name, sources=[])


class CMakeBuild(build_ext):

    def run(self):
        global binary_files

        is_win = platform.system() == 'Windows'

        try:
            _ = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        # preparing cmake arguments

        cmake_args = [
            ('-DCMAKE_LIBRARY_OUTPUT_DIRECTORY='
                + os.path.abspath(self.build_temp)),
            '-DPYTHON_EXECUTABLE=' + sys.executable,
            '-DCMAKE_CXX_STANDARD=17',
        ]

        # determining build type

        cfg = 'Debug' if self.debug else 'Release'
        self.build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        # parallelize compilation, if using make files

        if not is_win:
            cpu_count = multiprocessing.cpu_count()
            self.build_args += ['--', '-j{}'.format(cpu_count)]

        # additional flags

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
                                env.get('CXXFLAGS', ''),
                                self.distribution.get_version())

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # call cmake to configure and build

        cmake_dir = os.path.abspath(os.path.dirname(__file__))

        subprocess.check_call(['cmake', cmake_dir] + cmake_args,
                              cwd=self.build_temp,
                              env=env)

        cmake_cmd = ['cmake', '--build', '.'] + self.build_args

        subprocess.check_call(cmake_cmd, cwd=self.build_temp)

        # Are we installing or building wheel?
        is_wheel = 'bdist_wheel' in sys.argv

        # Detect binary files to include in wheel
        for ext in self.extensions:
            build_temp = Path(self.build_temp).resolve()
            if platform.system() == 'Windows':
                build_temp = build_temp / 'Release'

            suff = Path(self.get_ext_filename(ext.name)).suffix
            fmts = set([suff, '.dll', '.pyd', '.so'])
            matches = [p for p in build_temp.glob('*') if p.suffix in fmts]
            dest_dir = Path(self.get_ext_fullpath(ext.name)).resolve().parent

            for m in matches:
                binary_files.append(m)
                if not is_wheel:
                    self.copy_file(m, dest_dir / m.name) # tmp => lib


setup(name="imviz",
      version="0.2.3",
      description="Pythonic bindings for imgui/implot",
      url="https://github.com/joruof/imviz",
      author="Jona Ruof",
      author_email="jona.ruof@uni-ulm.de",
      license="MIT",
      zip_safe=False,
      long_description=Path('README.md').read_text('utf-8'),
      long_description_content_type="text/markdown",
      python_requires=">=3.6",
      packages=find_packages(),
      ext_modules=[CMakeExtension("cppimviz")],
      cmdclass=dict(build_ext=CMakeBuild, install=Install),
      include_package_data=True,
      install_requires=[
            "numpy", "Pillow>=9.0.1", "objtoolbox>=0.0.9", "minireload>=0.0.5"
          ]
      )
